/* $Id$ */

/** @file date_gui.cpp */

#include "stdafx.h"
#include "window_func.h"
#include "core/math_func.hpp"
#include "strings_type.h"
#include "strings_func.h"
#include "string_type.h"
#include "date_func.h"
#include "gfx_type.h"
#include "gfx_func.h"
#include "window_gui.h"
#include "querystring_gui.h"

#include "widgets/dropdown_type.h"
#include "widgets/dropdown_func.h"

#include "table/strings.h"

enum SetDateWidgets {
	SDW_CLOSEBOX = 0,
	SDW_CAPTION,
	SDW_BACKGROUND,
	SDW_DAY,
	SDW_MONTH,
	SDW_YEAR,
	SDW_SET_DATE,
};

struct SetDateWindow : Window {
	void (*proc)(Window*, uint32);
	YearMonthDay date;
	Year min_year;
	Year max_year;

	SetDateWindow(const WindowDesc *desc, WindowNumber window_number, Window *parent, Date initial_date, Year min_year, Year max_year, void (*callback)(Window*, uint32)) : Window(desc, window_number)
	{
		assert(min_year <= max_year);

		this->parent = parent;
		this->left = parent->left + (parent->width / 2) - (this->width / 2);
		this->top = parent->top + (parent->height / 2) - (this->height / 2);

		this->proc = callback;
		this->min_year = min_year;
		this->max_year = max_year;

		if (initial_date == 0) initial_date = _date;
		ConvertDateToYMD(initial_date, &this->date);

		this->date.year = Clamp(this->date.year, min_year, max_year);
	}

	void ShowDateDropDown(int type)
	{
		int selected;
		DropDownList *list = new DropDownList();

		switch (type) {
			default: NOT_REACHED();

			case SDW_DAY:
				for (uint i = 0; i <= 30; i++) {
					list->push_back(new DropDownListStringItem(STR_ORDINAL_NUMBER_1ST + i, i + 1, false));
				}
				selected = this->date.day;
				break;

			case SDW_MONTH:
				for (uint i = 0; i <= 11; i++) {
					list->push_back(new DropDownListStringItem(STR_MONTH_JAN + i, i, false));
				}
				selected = this->date.month;
				break;

			case SDW_YEAR:
				for (Year i = this->min_year; i <= this->max_year; i++) {
					DropDownListParamStringItem *item = new DropDownListParamStringItem(STR_JUST_INT, i, false);
					item->SetParam(0, i);
					list->push_back(item);
				}
				selected = this->date.year;
				break;
		}

		ShowDropDownList(this, list, selected, type);
	}

	virtual void OnPaint()
	{
		this->widget[SDW_DAY].data   = STR_ORDINAL_NUMBER_1ST + this->date.day - 1;
		this->widget[SDW_MONTH].data = STR_MONTH_JAN + this->date.month;
		SetDParam(0, this->date.year);

		this->DrawWidgets();
	}

	virtual void OnClick(Point pt, int widget)
	{
		switch (widget) {
			case SDW_DAY:
			case SDW_MONTH:
			case SDW_YEAR:
				ShowDateDropDown(widget);
				break;

			case SDW_SET_DATE:
				if (this->proc != NULL) this->proc(this->parent, ConvertYMDToDate(this->date.year, this->date.month, this->date.day));
				delete this;
				break;
		}
	}

	virtual void OnDropdownSelect(int widget, int index)
	{
		switch (widget) {
			case SDW_DAY:
				this->date.day = index;
				break;

			case SDW_MONTH:
				this->date.month = index;
				break;

			case SDW_YEAR:
				this->date.year = index;
				break;
		}
		this->SetDirty();
	}
};

static const Widget _set_date_widgets[] = {
	{   WWT_CLOSEBOX,   RESIZE_NONE,  COLOUR_BROWN,     0,    10,     0,    13, STR_BLACK_CROSS, STR_TOOLTIP_CLOSE_WINDOW},             // SDW_CLOSEBOX
	{    WWT_CAPTION,   RESIZE_NONE,  COLOUR_BROWN,    11,   199,     0,    13, STR_SET_DATE,    STR_TOOLTIP_WINDOW_TITLE_DRAG_THIS},   // SDW_CAPTION
	{      WWT_PANEL,   RESIZE_NONE,  COLOUR_BROWN,     0,   199,    14,    43, 0x0,             STR_NULL},                          // SDW_BACKGROUND
	{   WWT_DROPDOWN,   RESIZE_NONE, COLOUR_ORANGE,    10,    50,    24,    35, STR_NULL,        STR_DAY_TOOLTIP},                   // SDW_DAY
	{   WWT_DROPDOWN,   RESIZE_NONE, COLOUR_ORANGE,    60,   130,    24,    35, STR_NULL,        STR_MONTH_TOOLTIP},                 // SDW_MONTH
	{   WWT_DROPDOWN,   RESIZE_NONE, COLOUR_ORANGE,   140,   190,    24,    35, STR_JUST_INT,    STR_YEAR_TOOLTIP},                  // SDW_YEAR
	{ WWT_PUSHTXTBTN,   RESIZE_NONE,  COLOUR_BROWN,     0,   199,    44,    55, STR_SET_DATE,    STR_NULL},                          // SDW_SET_DATE

	{    WIDGETS_END }
};

static const WindowDesc _set_date_desc(
	WDP_CENTER, WDP_CENTER, 200, 56, 200, 56,
	WC_SET_DATE, WC_NONE,
	WDF_STD_TOOLTIPS | WDF_STD_BTN | WDF_DEF_WIDGET | WDF_UNCLICK_BUTTONS,
	_set_date_widgets
);

void ShowSetDateWindow(Window *parent, int window_number, Date initial_date, Year min_year, Year max_year, void (*callback)(Window*, uint32))
{
	DeleteWindowByClass(WC_SET_DATE);
	new SetDateWindow(&_set_date_desc, window_number, parent, initial_date, min_year, max_year, callback);

}


enum SetDateHHMMSSWidgets {
	SDHW_CLOSEBOX = 0,
	SDHW_CAPTION,
	SDHW_BACKGROUND,
	SDHW_TIME_LABEL,
	SDHW_TIME_EDITBOX,
	SDHW_HOUR,
	SDHW_MINUTE,
	SDHW_SECOND,
	SDHW_SET_TIME,
};

struct SetDateHHMMSSWindow : QueryStringBaseWindow {
	void (*proc)(Window*, uint32);
	HourMinSec time;
	bool show_seconds;

	SetDateHHMMSSWindow(const WindowDesc *desc, WindowNumber window_number, Window *parent, Time initial_time, bool seconds, void (*callback)(Window*, uint32)) : QueryStringBaseWindow(9, desc, window_number)
	{
		this->parent = parent;
		this->left = parent->left + (parent->width / 2) - (this->width / 2);
		this->top = parent->top + (parent->height / 2) - (this->height / 2);

		this->proc = callback;

		if (initial_time == INVALID_TIME) initial_time = _virtual_time;
		ConvertTimeToHMS(initial_time, &this->time);
		if (!seconds) this->time.sec = 0; // round to minutes when seconds not wanted

		show_seconds = seconds;

		snprintf(this->edit_str_buf, this->edit_str_size, show_seconds ? "%02d:%02d:%02d" : "%02d:%02d", this->time.hour, this->time.min, this->time.sec);
		this->afilter = CS_NUM_OR_COLON;
		InitializeTextBuffer(&this->text, this->edit_str_buf, this->edit_str_size, 90);
	}

	void ShowDateDropDown(int type)
	{
		int selected;
		DropDownList *list = new DropDownList();

		switch (type) {
			default: NOT_REACHED();

			case SDHW_HOUR:
				for (uint i = 0; i < 24; i++) {
					DropDownListParamStringItem *item = new DropDownListParamStringItem(STR_JUST_INT, i, false);
					item->SetParam(0, i);
					list->push_back(item);
				}
				selected = this->time.hour;
				break;

			case SDHW_MINUTE:
			case SDHW_SECOND:
				for (uint i = 0; i < 60; i++) {
					DropDownListParamStringItem *item = new DropDownListParamStringItem(STR_JUST_INT, i, false);
					item->SetParam(0, i);
					list->push_back(item);
				}
				selected = ((type == SDHW_MINUTE) ? this->time.min : this->time.sec);
				break;
		}

		ShowDropDownList(this, list, selected, type);
	}

	void UpdateEditBox()
	{
		snprintf(this->text.buf, this->text.maxsize, show_seconds ? "%02d:%02d:%02d" : "%02d:%02d", this->time.hour, this->time.min, this->time.sec);
		UpdateTextBufferSize(&this->text);
		this->SetWidgetDirty(SDHW_TIME_EDITBOX);
	}

	void RefreshTimeFromEditBox()
	{
		unsigned h, m, s;
		char trail; // This is to determine if there were additional characters

		s = 0;
		int assigned = sscanf(this->text.buf, "%2u:%2u:%2u%c", &h, &m, &s, &trail);

		if (assigned == 2 || assigned == 3) {
			this->time.hour = h;
			this->time.min = m;
			this->time.sec = s;
		}
	}

	virtual void OnPaint()
	{
		SetDParam(0, this->time.hour);
		SetDParam(1, this->time.min);
		SetDParam(2, this->time.sec);

		SetWidgetDisabledState(SDHW_SECOND, !show_seconds);

		this->DrawWidgets();
		this->DrawEditBox(SDHW_TIME_EDITBOX);
	}

	virtual void OnClick(Point pt, int widget)
	{
		switch (widget) {
			case SDHW_TIME_EDITBOX:
				ShowOnScreenKeyboard(this, widget, 0, 0);
				break;

			case SDHW_HOUR:
			case SDHW_MINUTE:
			case SDHW_SECOND:
				ShowDateDropDown(widget);
				break;

			case SDHW_SET_TIME:
				if (this->proc != NULL) this->proc(this->parent, ConvertHMSToTime(&this->time));
				delete this;
				break;
		}
	}

	virtual void OnDropdownSelect(int widget, int index)
	{
		switch (widget) {
			case SDHW_HOUR:
				this->time.hour = index;
				break;

			case SDHW_MINUTE:
				this->time.min = index;
				break;

			case SDHW_SECOND:
				this->time.sec = index;
				break;
		}
		this->UpdateEditBox();
		this->SetDirty();
	}

	virtual void OnMouseLoop()
	{
		this->HandleEditBox(SDHW_TIME_EDITBOX);
	}

	virtual EventState OnKeyPress(uint16 key, uint16 keycode)
	{
		if (keycode == WKC_ESC) {
			delete this;
			return ES_HANDLED;
		}

		EventState state = ES_NOT_HANDLED;
		if (this->HandleEditBoxKey(SDHW_TIME_EDITBOX, key, keycode, state) == 1) { // Press Enter
			Point pt = { 0, 0 };
			this->HandleButtonClick(SDHW_SET_TIME);
			this->OnClick(pt, SDHW_SET_TIME);
		} else if (state == ES_HANDLED) { // Handled and not Enter
			RefreshTimeFromEditBox();
			this->SetWidgetDirty(SDHW_HOUR);
			this->SetWidgetDirty(SDHW_MINUTE);
			this->SetWidgetDirty(SDHW_SECOND);
		}

		return state;
	}
};

static const Widget _set_date_hhmmss_widgets[] = {
	{   WWT_CLOSEBOX,   RESIZE_NONE,    COLOUR_BROWN,     0,    10,     0,    13, STR_BLACK_CROSS, STR_TOOLTIP_CLOSE_WINDOW},             // SDHW_CLOSEBOX
	{    WWT_CAPTION,   RESIZE_NONE,    COLOUR_BROWN,    11,   159,     0,    13, STR_SET_TIME,    STR_TOOLTIP_WINDOW_TITLE_DRAG_THIS},   // SDHW_CAPTION
	{      WWT_PANEL,   RESIZE_NONE,    COLOUR_BROWN,     0,   159,    14,    63, 0x0,             STR_NULL},                          // SDHW_BACKGROUND
	{       WWT_TEXT,   RESIZE_NONE,    COLOUR_WHITE,    10,    50,    24,    35, STR_TIME,        STR_NULL},                          // SDHW_TIME_LABEL
	{    WWT_EDITBOX,   RESIZE_NONE,    COLOUR_BROWN,    60,   150,    24,    35, STR_SET_TIME,    STR_ENTER_TIME_TOOLTIP},            // SDHW_TIME_EDITBOX
	{   WWT_DROPDOWN,   RESIZE_NONE,   COLOUR_ORANGE,    10,    50,    44,    55, STR_JUST_INT,    STR_HOUR_TOOLTIP},                  // SDHW_HOUR
	{   WWT_DROPDOWN,   RESIZE_NONE,   COLOUR_ORANGE,    60,   100,    44,    55, STR_NUM_1,       STR_MIN_TOOLTIP},                   // SDHW_MINUTE
	{   WWT_DROPDOWN,   RESIZE_NONE,   COLOUR_ORANGE,   110,   150,    44,    55, STR_NUM_2,       STR_SEC_TOOLTIP},                   // SDHW_SECOND
	{ WWT_PUSHTXTBTN,   RESIZE_NONE,    COLOUR_BROWN,     0,   159,    64,    75, STR_SET_TIME,    STR_NULL},                          // SDHW_SET_TIME

	{    WIDGETS_END }
};

static const WindowDesc _set_date_hhmmss_desc(
	WDP_CENTER, WDP_CENTER, 160, 76, 160, 76,
	WC_SET_DATE, WC_NONE,
	WDF_STD_TOOLTIPS | WDF_STD_BTN | WDF_DEF_WIDGET | WDF_UNCLICK_BUTTONS,
	_set_date_hhmmss_widgets
);

void ShowSetDateHHMMSSWindow(Window *parent, int window_number, Time initial_time, bool seconds, void (*callback)(Window*, uint32))
{
	DeleteWindowById(WC_SET_DATE, window_number);
	new SetDateHHMMSSWindow(&_set_date_hhmmss_desc, window_number, parent, initial_time, seconds, callback);
}
