/* $Id$ */

/** @file headway_cmd.cpp Commands and OrderList methods related to headway. */

#include "stdafx.h"
#include "functions.h"
#include "settings_type.h"
#include "command_func.h"
#include "strings_func.h"
#include "window_type.h"
#include "window_func.h"
#include "date_func.h"
#include "vehicle_base.h"
#include "order_base.h"
#include "news_type.h"
#include "news_func.h"
#include "company_func.h"
#include "debug.h"

#include "table/strings.h"

/**
 * Start headway i.e. automatically separating the shared vehicles in the
 * timetable according to the given interval
 * @param tile Not used.
 * @param flags Operation to perform.
 * @param p1 Various bitstuffed elements
 * - p1 = (bit  0 - 15) - OrderListID
 * - p1 = (bit 16 - 31) - VehicleID of start vehicle (INVALID_VEHICLE for next).
 * @param p2 Various bitstuffed elements
 * - p2 = (bit  0 - 19) - interval between start dates in ticks (in case of
 *                        autofill this may be all ones to autodetect)
 * - p2 = (bit 20 - 31) - unused
 */
CommandCost CmdSetHeadwayInterval(TileIndex tile, DoCommandFlag flags, uint32 p1, uint32 p2, const char *text)
{
	if (!_settings_game.order.timetabling) return CMD_ERROR;

	OrderListID ol = GB(p1, 0, 16);
	OrderList *orders = OrderList::GetIfValid(ol);
	if (orders == NULL) return CMD_ERROR;

	VehicleID veh = GB(p1, 16, 16);
	bool use_v_as_leader = Vehicle::IsValidID(veh);
	Vehicle *v = Vehicle::GetIfValid(veh);
	if (v == NULL) v = orders->GetFirstSharedVehicle();
	if (v == NULL) return CMD_ERROR;
	if (!CheckOwnership(v->owner)) return CMD_ERROR;

	if (v->orders.list != orders) return CMD_ERROR;

	Ticks interval = (Ticks)GB(p2, 0, 20);
	bool disable = (interval == 0);
	bool autofill = use_v_as_leader && HasBit(v->vehicle_flags, VF_AUTOFILL_TIMETABLE);

	if (!disable) {
		if (!v->IsOrderListShared() || v->GetNumOrders() == 0) return CMD_ERROR;
		if (use_v_as_leader && HasBit(v->vehstatus, VS_STOPPED)) return CMD_ERROR;
		if (!autofill && v->GetTimetableTotalDuration() == INVALID_TICKS) return CMD_ERROR;

		/* We need at least two vehicles and at most as many as fit into the timetable
		 * without overlap */
		int count = v->orders.list->GetNumRunningVehicles();
		if (count < 2 || (!autofill && (count - 1)*interval >= v->GetTimetableTotalDuration())) return CMD_ERROR;
	}

	if (flags & DC_EXEC) {
		if (disable) {
			orders->DisableHeadway();
		}
		else {
			if (autofill && interval == (1 << 20) - 1) interval = INVALID_TICKS;
			orders->EnableHeadway(interval, use_v_as_leader ? v : NULL, autofill);
		}
	}

	return CommandCost();
}

/**
 * Set the headway to automatic maintenance mode.
 * @param tile Not used.
 * @param flags Operation to perform.
 * @param p1 Various bitstuffed elements
 * - p1 = (bit  0 - 15) - OrderListID
 * - p1 = (bit 31)      - 1=enable, 0=disable
 * @param p2 Not used
 */
CommandCost CmdSetHeadwayAutomatic(TileIndex tile, DoCommandFlag flags, uint32 p1, uint32 p2, const char *text)
{
	if (!_settings_game.order.timetabling) return CMD_ERROR;

	OrderListID ol = GB(p1, 0, 16);
	OrderList *orders = OrderList::GetIfValid(ol);
	if (orders == NULL) return CMD_ERROR;

	Vehicle *v = orders->GetFirstSharedVehicle();
	if (v == NULL) return CMD_ERROR;
	if (!CheckOwnership(v->owner)) return CMD_ERROR;

	if ((orders->GetHeadwayFlags() & (HWFB_ACTIVE | HWFB_INITIALIZE)) == 0) return CMD_ERROR;

	bool enable = HasBit(p1, 31);
	if (enable) {
		int count = orders->GetNumRunningVehicles();
		if (count < 2 || (count - 1)*orders->GetHeadwayInterval() >= v->GetTimetableTotalDuration()) return CMD_ERROR;
	}

	if (flags & DC_EXEC) {
		orders->SetHeadwayAutomaticMode(enable);
	}

	return CommandCost();
}

/**
 * Correct lateness of all vehicles of the headway by a given amount.
 * @param tile Not used.
 * @param flags Operation to perform.
 * @param p1 Various bitstuffed elements
 * - p1 = (bit  0 - 15) - OrderListID
 * - p1 = (bit 16)      - automatically determine offset to make all vehicles on time
 * - p1 = (bit 17)      - automatically determine offset to make at least one vehicle on time
 * - p1 = (bit 18)      - round to multiple of interval (towards minus infinity)
 * - p1 = (bit 19)      - never make vehicles more late
 * @param p2 offset in one-complement, i.e.
 * - p2 = (bit  0 - 30) - absolute value of offset
 * - p2 = (bit 31)      - sign of offset (1 = negative)
 */
CommandCost CmdSetHeadwayLateness(TileIndex tile, DoCommandFlag flags, uint32 p1, uint32 p2, const char *text)
{
	if (!_settings_game.order.timetabling) return CMD_ERROR;

	OrderListID ol = GB(p1, 0, 16);
	OrderList *orders = OrderList::GetIfValid(ol);
	if (orders == NULL) return CMD_ERROR;

	Vehicle *v = orders->GetFirstSharedVehicle();
	if (v == NULL) return CMD_ERROR;
	if (!CheckOwnership(v->owner)) return CMD_ERROR;

	if ((orders->GetHeadwayFlags() & (HWFB_ACTIVE | HWFB_INITIALIZE)) == 0) return CMD_ERROR;
	/* Messing around with lateness now could confuse headway */
	if (HasBit(orders->GetHeadwayFlags(), HWF_FIRST_ROUND)) return CMD_ERROR;

	bool automatic_max = HasBit(p1, 16);
	bool automatic_min = HasBit(p1, 17);
	bool round = HasBit(p1, 18);

	/* Not both! */
	if (automatic_max && automatic_min) return CMD_ERROR;
	/* Make up your mind, automatic or manual? */
	if ((automatic_max || automatic_min) && p2 != 0) return CMD_ERROR;

	int32 max_lateness;
	int32 min_lateness;
	orders->CalculateLatenessStatistics(&min_lateness, &max_lateness);

	bool negative;
	int64 offset;

	if (automatic_max) {
		negative = -max_lateness < 0;
		offset = abs(-max_lateness);
	} else if (automatic_min) {
		negative = -min_lateness < 0;
		offset = abs(-min_lateness);
	} else {
		negative = HasBit(p2, 31);
		offset = GB(p2, 0, 31);
		if (offset == 0) return CMD_ERROR; // this gets rid of -0 also
	}

	/* Do the rounding before negating as rounding behavior of dividing negative
	 * values is implementation-defined by ISO-C++ standard. */
	if (round) {
		if (negative) offset += orders->GetHeadwayInterval() - 1;
		else offset -= orders->GetHeadwayInterval() - 1;
		offset -= offset % orders->GetHeadwayInterval();
	}
	if (negative) offset = -offset;
	if ((int64)min_lateness + offset < INT32_MIN || (int64)max_lateness + offset > INT32_MAX) return CMD_ERROR;

	if (offset > 0 && HasBit(p2, 19)) return CommandCost();

	if (flags & DC_EXEC) {
		for (; v != NULL; v = v->NextShared()) {
			v->lateness_counter += offset;
		}
		orders->AdjustHeadwayStart(offset);
	}

	return CommandCost();
}


void OrderList::EnableHeadway(Ticks interval, Vehicle *leader, bool autofill)
{
	this->headway_interval = interval;
	this->headway_start_vehicle = leader;

	if (autofill) {
		SetBit(this->headway_flags, HWF_AUTOFILL);
		if (HasBit(leader->vehicle_flags, VF_TIMETABLE_STARTED)) SetBit(this->headway_flags, HWF_INITIALIZE);
		else ClrBit(this->headway_flags, HWF_INITIALIZE);
	}
	else {
		ClrBit(this->headway_flags, HWF_AUTOFILL);
		SetBit(this->headway_flags, HWF_INITIALIZE);
	}

	ClrBit(this->headway_flags, HWF_FIRST_ROUND);
}

void OrderList::DisableHeadway(bool keep_interval)
{
	this->headway_flags = 0;
	if (!keep_interval) this->headway_interval = INVALID_TICKS;
	this->headway_last_start = INVALID_TICKS;
	this->headway_last_vehicle = NULL;
	this->headway_start_vehicle = NULL;
}

void OrderList::SetHeadwayAutomaticMode(bool enable)
{
	assert(this->headway_flags & (HWFB_ACTIVE | HWFB_INITIALIZE | HWFB_AUTOFILL));

	if (enable) SetBit(this->headway_flags, HWF_AUTOMATIC);
	else ClrBit(this->headway_flags, HWF_AUTOMATIC);
}

void OrderList::UpdateHeadway(Vehicle *v, bool travelling)
{
	/* The headway list may be open when headway is disabled */
	InvalidateWindowData(WC_HEADWAY_LIST, this->index, 0);

	/* No headway */
	if (this->headway_flags == 0) return;

	assert(this->num_running_veh >= 2);

	/* We do the rest of our business only at the start of a timetable */
	if (v->cur_order_index != 0 || !travelling) return;

	bool disable = false;

	if (HasBit(this->headway_flags, HWF_AUTOFILL)) {
		assert(this->headway_start_vehicle != NULL);

		if (this->headway_start_vehicle == v) {
			if (!HasBit(this->headway_flags, HWF_INITIALIZE)) {
				/*  Start of the autofill round */
				DEBUG(tt, 3, "Headway leader %s %hu starts autofill", v->GetTypeString(), v->unitnumber);

				SetBit(this->headway_flags, HWF_INITIALIZE);
				return;
			} else {
				/* Autofill has completed, check or calculate interval */
				if (this->headway_interval == INVALID_TICKS) {
					this->headway_interval = this->timetable_duration / this->num_running_veh;
				}

				disable = (this->num_running_veh < 2 || ((int)this->num_running_veh - 1)*this->headway_interval >= this->timetable_duration);

				ClrBit(this->headway_flags, HWF_AUTOFILL);
			}
		}	else {
			DEBUG(tt, 4, "Passing %s %hu is not headway leader during headway autofill", v->GetTypeString(), v->unitnumber);
			return;
		}
	}

	disable |= !this->IsCompleteTimetable();

	if (disable) {
		DEBUG(tt, 1, "Headway disabled due to incomplete timetable for %s %hu", v->GetTypeString(), v->unitnumber);

		DisableHeadway(true);
		if (v->owner == _local_company) {
			/* Notify the user that we stopped the headway */
			SetDParam(0, v->index);
			AddNewsItem(STR_HEADWAY_FAILURE, NS_ADVICE, NR_VEHICLE, v->index);
		}
		return;
	}

	if (HasBit(this->headway_flags, HWF_INITIALIZE)) {
		assert(!HasBit(this->headway_flags, HWF_FIRST_ROUND));

		if (this->headway_start_vehicle == NULL || this->headway_start_vehicle == v) {
			/* Start first round of headway if this is the desired leader or if the
			 * headway should just start with the next vehicle
			 */
			assert(v->timetable_start == 0); // Should already be handled

			DEBUG(tt, 3, "Headway leader %s %hu starts first round", v->GetTypeString(), v->unitnumber);

			ClrBit(this->headway_flags, HWF_INITIALIZE);
			SetBit(this->headway_flags, HWF_ACTIVE);
			SetBit(this->headway_flags, HWF_FIRST_ROUND);
			this->headway_start_vehicle = v; // to detect end of first round
		} else {
			DEBUG(tt, 4, "Passing %s %hu is not headway leader during headway initialization", v->GetTypeString(), v->unitnumber);
			return;
		}
	} else if (HasBit(this->headway_flags, HWF_FIRST_ROUND)) {
		assert(HasBit(this->headway_flags, HWF_ACTIVE));
		assert(!HasBit(this->headway_flags, HWF_INITIALIZE));
		assert(this->headway_start_vehicle != NULL);

		if (this->headway_start_vehicle == v) {
			/* First round finished */
			DEBUG(tt, 3, "Headway leader %s %hu finishes first round", v->GetTypeString(), v->unitnumber);
			ClrBit(this->headway_flags, HWF_FIRST_ROUND);
		} else  {
			/* During first round */
			DEBUG(tt, 4, "Integrate %s %hu into headway during first round", v->GetTypeString(), v->unitnumber);
			assert(this->headway_last_vehicle != NULL);
			assert(this->headway_last_start != INVALID_TICKS);

			Ticks intended_start = this->headway_last_start + this->headway_interval;
			v->lateness_counter = DateToTicks() - intended_start;
		}
	} else if (HasBit(this->headway_flags, HWF_AUTOMATIC)) {
		/* Automatic mode after first round */
		assert(this->headway_last_start != INVALID_TICKS);

		DEBUG(tt, 4, "Do automatic headway maintenance for %s %hu ", v->GetTypeString(), v->unitnumber);

		/* We need at least two vehicles and at most as many as fit into the timetable
		 * without overlap */
		int count = this->num_running_veh;
		if (!this->IsCompleteTimetable() || count < 2 || (count - 1)*this->headway_interval >= this->timetable_duration) {
			DEBUG(tt, 1, "Automatic headway maintenance disables itself for %s %hu", v->GetTypeString(), v->unitnumber);

			ClrBit(this->headway_flags, HWF_AUTOMATIC);
			if (v->owner == _local_company) {
				/* Notify the user that we stopped the headway */
				SetDParam(0, v->index);
				AddNewsItem(STR_AUTOMATIC_HEADWAY_FAILURE, NS_ADVICE, NR_VEHICLE, v->index);
			}
			return;
		}

		Ticks interval;
		if (this->headway_start_vehicle == NULL || this->headway_start_vehicle == v) {
			interval = this->timetable_duration - (count - 1) * this->headway_interval;
		} else {
			interval = this->headway_interval;
		}
		Ticks intended_start = this->headway_last_start + interval;
		Ticks new_lateness = DateToTicks() - intended_start;
		if (new_lateness != v->lateness_counter) {
			DEBUG(tt, 3, "Automatic headway maintenance corrects lateness for %s %hu from %d to %d", v->GetTypeString(), v->unitnumber, v->lateness_counter, new_lateness);
			v->lateness_counter = new_lateness;
		}
	}

	/* If the leader was removed, make this vehicle the leader */
	if (this->headway_start_vehicle == NULL) {
		DEBUG(tt, 3, "Elected %s %hu as new headway leader", v->GetTypeString(), v->unitnumber);
		this->headway_start_vehicle = v;
	}

	this->headway_last_start = DateToTicks() - v->lateness_counter;
	this->headway_last_vehicle = v;
}
